#include <nspr.h>
#include <math.h>
#include <limits.h>
/* Table of constant values */

/* sqrt(2) */
#define SQ2 1.41421356237310

/* Computes integral of functions.
 * Functions may have singularities at one or both end-points of an interval
 *  (a,b), see [1, 2].
 * @param a
 *   The lower bound for integration, or -inf
 * @param b
 *   The upper bound for integration, or +inf
 * @param relerr
 *   The relative error bound.
 * @param maxsteps
 *   The maximum number of evaluation steps allowed.
 * 
 * References:
 * [1] Sikorski, K., Optimal quadrature algorithms in HP
 *     spaces, Num. Math., 39, 405-410 (1982).
 * [2] Sikorski, K., Stenger, F., Optimal quadratures in
 *     HP spaces, ACM TOMS.
 */

enum
{
  DI_OUTER_LOOP,
  DI_INNER_LOOP,
  DI_FINISHED,
  DI_FAILED
} di_state;

/*
 * 1: Infinite at both ends.
 * 2: Open at one end(not used). The literature indicates that this is best
 *      for non-oscillatory functions, but don't have this information, so this
 *      code is currently unused(but preserved).
 * 3: Open at one end(best for oscillatory functions).
 * 4: Finite at both ends
 */
PRUint32 di_boundtype;

double di_exph, di_h, di_exph0, di_h0, di_h1, di_e1, di_cor, di_u, di_sum;
double di_a, di_b, di_sr, di_xmultiply, di_ba, di_eps3, di_v0, di_v1, di_v2;
double di_relerr, di_ret, *di_mainbound, *di_allvars, di_sum2, di_s;
double *di_output;
PRBool di_inf1, di_inf2;
PRInt32 di_i, di_m2, di_i1, di_n1, di_k, di_l, di_m1, di_boundvar;
PRInt32 di_l1;

#define CALL_DI_FUNCTION(x) \
  {\
    double backup;\
    if (di_boundvar == -1) \
    {\
      backup = *di_mainbound; \
      *di_mainbound = di_xmultiply * x; \
    }\
    else if (di_boundvar < -1)\
    {\
      backup = constants[-2 - di_boundvar];\
      constants[-2 - di_boundvar] = di_xmultiply * x; \
    }\
    else \
    {\
      backup = di_allvars[di_boundvar]; \
      di_allvars[di_boundvar] = di_xmultiply * x; \
    }\
    di_ret = (di_function)(di_boundvar, di_allvars); \
    if (di_boundvar == -1) \
      *di_mainbound = backup; \
    else if (di_boundvar < -1) \
      constants[-2 - di_boundvar] = backup; \
    else \
      di_allvars[di_boundvar] = backup; \
  }

double (*di_function)(double, double*);

static inline PRBool is_there_input();

void
BeginDIInnerLoop()
{
  double w1, w2, v, sum1;
  switch (di_boundtype)
  {
  case 1:
    CALL_DI_FUNCTION(di_h1);
    v = di_ret;
    di_h1 += di_h;
    break;
  case 2:
    CALL_DI_FUNCTION((di_a + di_e1));
    v = di_e1 * di_ret;
    di_e1 *= di_exph;
    break;
  case 3:
    w1 = sqrt(di_e1 + 1.0 / di_e1);
    w2 = sqrt(di_e1);
    if (di_e1 < 0.1)
    {
      double c0, s1, t, w3, w4;
      w3 = di_e1;
      w4 = di_e1 * di_e1;
      c0 = 1.0;
      di_s = di_e1;
      s1 = di_e1;
      t = 0.0;
      for (t = 0.0;
           (t == 0.0) || (di_s != s1);
           s1 = di_s)
      {
        c0 = -c0 * (t + 0.5)
                 * (t * 2.0 + 1.0)
                 / (t * 2.0 + 3.0)
                 / (t + 1.0);
        t += 1.0;
        w3 *= w4;
        di_s += c0 * w3;
      }
    }
    else
      di_s = log(di_e1 + w1 * w2);
    CALL_DI_FUNCTION((di_a + di_s))
    v = w2 / w1 * di_ret;
    di_e1 *= di_exph;
    break;
  case 4:
    w1 = di_e1 + 1.0;
    CALL_DI_FUNCTION((di_a + di_b * di_e1) / w1)
    v = di_e1 / w1 / w1 * di_ret * di_ba;
    di_e1 *= di_exph;
    break;
  }
  di_i++;
  sum1 = di_u + v;
  if (abs_d(di_u) < abs_d(v))
    di_cor = di_u - (sum1 - v) + di_cor;
  else
    di_cor = v - (sum1 - di_u) + di_cor;

  di_u = sum1;

  if (di_i < di_l1)
    return;
  
  /* Check for heuristic termination... */
  if (di_inf1)
  {
    di_v0 = di_v1;
    di_v1 = di_v2;
    di_v2 = abs_d(v);
    if (di_v0 + di_v1 + di_v2 > di_eps3)
    {
      if (di_i < di_m2)
        return;
      di_n1 = 5;
    }
    if (di_inf2)
      di_k = di_i;
    if (!di_inf2)
      di_l = di_i;
    di_v1 = di_relerr * 10.0;
    di_v2 = di_v1;
    di_m2 = di_m1 - di_l;
    if (!di_inf2)
    {
      di_inf2 = PR_TRUE;
      di_i = 0;
      di_exph = 1.0 / di_exph;
      di_h0 = -di_h0;
      di_e1 = 1.0 / di_exph0;
      di_h1 = di_h0;
      di_h = -di_h;
      /* Stay in the inner loop state... */
      return;
    }

    /* n1=5 - truncation condition not satisfied */
    if (di_n1 == 5)
    {
      /* This is essentially a failure... */
      integration_failed("defIntEvalExceeded");
      di_state = DI_FAILED;
      return;
    }

    /* Truncation condition satisfied, sum2=trapezoidal approximation */
    di_sum2 = sum1 + di_cor + di_sum;
    di_m2 = (di_k + di_l) << 1;
    
    /* Check the number of function evaluations */
    if (di_m2 > di_m1)
    {
      /* XXX "it is likely that quadr approximates the integral if m is large"
       * Perhaps we could use this result.
       */
      /* integration_failed("defIntEvalExceeded");
         di_state = DI_FAILED;*/
      *di_output = di_sum2;
      di_state = DI_FINISHED;
      return;
    }
    /* Initialise iteration */
    di_inf1 = PR_FALSE;
    di_inf2 = PR_FALSE;
    di_l1 = di_l;
    di_i = 0;
    di_h = -di_h;
    di_h0 = di_h / 2.0;
    di_state = DI_OUTER_LOOP;
    return;
  }
  else if (di_inf2)
  {
    di_h = -di_h;
    sum1 = (sum1 + di_cor) * di_h;
    w1 = (sum1 + di_sum2) / 2.0;
    
    /* Termination condition */
    if (abs_d(sum1 - di_sum2) <= di_sr)
    {
      *di_output = w1;
      di_state = DI_FINISHED;
      return;
    }
    
    /* Set up data for the next iteration */
    di_m2 <<= 1;
    if (di_m2 > di_m1)
    {
      /* XXX "it is likely that quadr approximates the integral if m is large"
       * Perhaps we could use this result.
       */
      /*
      integration_failed("defIntEvalExceeded");
      di_state = DI_FAILED;
      */
      *di_output = w1;
      di_state = DI_FINISHED;
      return;
    }
    di_i = 0;
    di_k <<= 1;
    di_l <<= 1;
    di_l1 = di_l;
    di_h /= 2.0;
    di_h0 = di_h / 2.0;
    di_sum2 = w1;
    di_inf2 = PR_FALSE;
    di_state = DI_OUTER_LOOP;
    return;
  }
  di_l1 = di_k;
  di_inf2 = PR_TRUE;
  di_i = 0;
  di_exph = 1.0 / di_exph;
  di_h0 = -di_h0;
  di_e1 = 1.0 / di_exph0;
  di_h1 = di_h0;
  di_h = -di_h;
}

void
BeginDIOuterLoop()
{
  di_exph = exp(di_h);
  di_exph0 = exp(di_h0);
  di_h1 = di_h0;
  di_e1 = di_exph0;
  di_u = 0.0;
  di_cor = 0.0;
  di_state = DI_INNER_LOOP;
  BeginDIInnerLoop();
}

double
inthp(
      double a,
      double b,
      double (*f)(double, double*),
      PRUint32 maxsteps,
      double relerr,
      double* mainbound,
      double* allvars,
      PRInt32 whichBound
     )
{
  PRBool a_finite, b_finite;

  di_mainbound = mainbound;
  di_allvars = allvars;
  di_boundvar = whichBound;

  di_a = a;
  di_b = b;
  di_relerr = relerr;
  di_function = f;

  if (di_a > di_b)
  {
    double tmp = di_a;
    di_a = di_b;
    di_b = tmp;
  }

  a_finite = finite(di_a);
  b_finite = finite(di_b);

  di_xmultiply = 1.0;
  
  if (a_finite && b_finite)
    di_boundtype = 4;
  else if (a_finite && !b_finite)
    di_boundtype = 3;
  else if (!b_finite && a_finite)
  {
    di_xmultiply = -1.0;
    di_boundtype = 3;
    b = -a;
  }
  else
    di_boundtype = 1;
  
  di_i1 = di_boundtype - 2;
  di_ba = b - a;
  if (di_ba == 0)
    return 0.0;
  di_n1 = 0;
  
  if (relerr < __DBL_EPSILON__)
    relerr = __DBL_EPSILON__;
  
  /* Set up data for the heuristic termination */
  di_h = 1.0;
  di_h0 = 1.0;
  di_eps3 = di_relerr / 3.0;
  di_sr = sqrt(di_relerr);
  di_v1 = di_relerr * 10.0;
  di_v2 = di_v1;
  di_m1 = maxsteps - 1;
  di_m2 = di_m1 / 2 + (di_m1 & 1);
  di_l1 = 0;
  di_inf1 = PR_TRUE;
  di_inf2 = PR_FALSE;

  /* Initialize the quadrature */
  di_i = 0;
  switch (di_boundtype)
  {
  case 1:
    CALL_DI_FUNCTION(0.0);
    di_sum = di_ret;
    break;
  case 2:
    CALL_DI_FUNCTION(a + 1.0);
    di_sum = di_ret;
    break;
  case 3:
    CALL_DI_FUNCTION(a + log(SQ2 + 1.0));
    di_sum = di_ret / SQ2;
    break;
  case 4:
    CALL_DI_FUNCTION((di_a + di_b) / 2.0);
    di_sum = di_ret / 4.0 * di_ba;
  }

  /* For now, we don't unwind the entire stack and go back to main before
   * completing the function evaluation, instead we don't return until
   * we have evaluated the integral. I did this to avoid having to implement
   * a way to return from EvaluateSimple/EvaluateDifferentials and then
   * continue evaluation from where we left off.
   * This means we have to call the network code from here instead of from
   * main(), but at least we can save all of our state on the stack. This code
   * is nevertheless written as a state machine that uses globals, so it can
   * more easily be converted to the other model.
   */
  di_state = DI_OUTER_LOOP;
  
  double result = 0.0;
  di_output = &result;

  do
  {
    switch (di_state)
    {
    case DI_OUTER_LOOP:
      BeginDIOuterLoop();
      break;
    case DI_INNER_LOOP:
      BeginDIInnerLoop();
      break;
    case DI_FINISHED:
      ;
    }
    /* Now we need to check for any incoming packets... */
    if (is_there_input())
      return result;
  }
  while (di_state != DI_FINISHED &&
         di_state != DI_FAILED);

  return result;
}
