/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original developer)
 *
 * Portions of this code were adapted from
 * "Levenberg - Marquardt non-linear minimization algorithm", by Manolis
 * Lourakis <lourakis@ics.forth.gr>. These portions must be distributed under
 * the GPL and not the MPL.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <math.h>
#include <nspr.h>
#include <string.h>

#define DBL_MAX 1e+37
#define DBL_MIN 1e-37

enum
{
  SOLVE_FOR_E,
  COMPUTE_JACOBIAN,
  SOLVE_FOR_STEP,
  BEGIN_ITERATION
} solution_state;

#define CYCLE_FUNCTION do_fitter_cycle
#define integration_failed fitting_failed
#define IN_FITTER

/* Constant values controlling fitter... */
#define FITTER_MU 1E-03
#define BLOCKSZ 32
#define BLOCKSZ_SQ (BLOCKSZ * BLOCKSZ)
#define FITTER_STOP_THRESH 1E-13
#define FITTER_STOP_THRESH_SQ (FITTER_STOP_THRESH*FITTER_STOP_THRESH)
#define FITTER_DIFF_DELTA 1E-06
#define EPSILON 1E-12
#define EPSILON2 (EPSILON*EPSILON)

static void do_fitter_cycle();

#include "SpecialMaths.c"
#include "IntegratorCommand.h"
#include "IntegrationHelperHead.c"
#include "DefiniteIntegral.c"

static PRBool integrator_call;

static inline void ComputeDifferentials(double y, const double *x, double *dx);
static inline void EvaluateSimple(double y, double *x);
static void EvaluateFunction(double *params, double *imeasurements);


#include "FitterCommand.h"
#include <string.h>

static PRUint32 nmeasurements = 0;
static PRUint32 nparamxmeasurements, nparam2, fitter_domain_error;
struct FitterMeasurement *measurements;
static double start_bound;
static PRUint32 current_measurement_mark;
static double *current_measurements;

static double *fitter_e = NULL, /* nx1 */
              *fitter_hx, /* \hat{x}_i, nx1 */
              *fitter_jacTe, /* J^T e_i mx1 */
              *fitter_jac, /* nxm */
              *fitter_jacTjac, /* mxm */
              *fitter_Dp, /* mx1 */
              *fitter_diag_jacTjac, /* diagonal of J^T J, mx1 */
              *fitter_pDp, /* p + Dp, mx1 */
              *fitter_wrk, /* nx1 */
              *fitter_p;  /* mx1*/
static double fitter_p_eL2, fitter_jacTe_inf, fitter_pDp_eL2, fitter_p_L2,
              fitter_d, fitter_savep, fitter_Dp_L2, fitter_mu,
              abs_tolerance_sq;
static PRUint32 fitter_nu, fitter_updp, fitter_updjac, fitter_newjac,
                fitter_K, fitter_iteration, fitter_jac_iteration;

static PRUint32 nparameters;
/* Negative values are constants, positive are differentials.
 *
 * The differential part of all_y0values can be set once and are then left alone
 * so we don't need to store the values, only the indices...
 */
PRInt32 param_index[ND + NC];
static double all_y0values[NS + ND];

#define SET_PARAM(i,v) if (param_index[i]>=0) \
                         all_y0values[NS + param_index[i]] = v; \
                       else \
                         constants[-1 - param_index[i]] = v;
#define GET_PARAM(i) ((param_index[i]>=0) ? all_y0values[NS + param_index[i]] \
                                          : constants[-1 - param_index[i]])
static inline void
fitting_failed(const char* error)
{
  char *data;
  PRUint32 l;
  struct FitterReplyHeader *irh;
  l = sizeof(struct FitterReplyHeader) + strlen(error) + 1;
  data = (char*)PR_Malloc(l);
  helper_state = SUSPENDED;
  irh = (struct FitterReplyHeader*)data;
  irh->length = l;
  irh->reply = FITTING_FAILED;
  strcpy(data + sizeof(struct FitterReplyHeader), error);
  send_packet(data, l, PR_TRUE);
  if (fitter_e != NULL)
    PR_Free(fitter_e);
  fitter_e = NULL;
}

static PRInt32
set_measurement_size(PRUint32 newsize)
{
  if (newsize == 0)
  {
    if (nmeasurements != 0)
      PR_Free(measurements);
    nmeasurements = 0;
    return -1;
  }
  if (nmeasurements == 0)
  {
    measurements = PR_Malloc(newsize * sizeof(*measurements));
    if (measurements == NULL)
    {
      nmeasurements = 0;
      return 0;
    }
    nmeasurements = newsize;
    return -1;
  }
  if (nmeasurements == newsize)
    return -1;
  PR_Free(measurements);
  measurements = PR_Malloc(newsize * sizeof(*measurements));
  if (measurements == NULL)
  {
    nmeasurements = 0;
    return 0;
  }
  nmeasurements = newsize;
  return -1;
}

static void
SwapMeasurements(PRUint32 idx1, PRUint32 idx2)
{
  struct FitterMeasurement tmp;
  memcpy(&tmp, measurements + idx2, sizeof(struct FitterMeasurement));
  memcpy(measurements + idx2, measurements + idx1, sizeof(struct FitterMeasurement));
  memcpy(measurements + idx1, &tmp, sizeof(struct FitterMeasurement));
}

static void
SortMeasurements(PRUint32 idx1, PRUint32 idx2)
{
  PRUint32 mid, i;
  double midBound;
  if (idx1 >= idx2)
    return;
  if (idx2 - idx1 == 1)
  {
    if (measurements[idx2].bound >= measurements[idx1].bound)
      return;
    SwapMeasurements(idx1, idx2);
    return;
  }
  /* Hopefully there aren't 2^31+ measurements, but we could still check... */
  mid = (idx1 + idx2) >> 1;
  midBound = measurements[mid].bound;
  SwapMeasurements(idx2, mid);
  mid = idx1;
  for (i = idx1; i < idx2; i++)
    if (measurements[i].bound <= midBound)
      SwapMeasurements(i, mid++);
  SwapMeasurements(mid, idx2);
  if (mid > 0)
    SortMeasurements(idx1, mid - 1);
  SortMeasurements(mid + 1, idx2);
}

static void
cmd_start(void *data, PRUint32 length)
{
  struct StartFittingCommand *cmd;
  PRInt32 *fp;
  struct FitterMeasurement *fm;
  PRUint32 i;

  helper_state = RUNNING;

  cmd = (struct StartFittingCommand*)data;
  abs_tolerance = cmd->abs_tolerance;
  abs_tolerance_sq = abs_tolerance * abs_tolerance;
  rel_tolerance = cmd->rel_tolerance;
  start_bound = cmd->start_bound;
  memcpy(
         all_y0values,
         cmd->initialValues,
         sizeof(cmd->initialValues)
        );
  memcpy(
         constants,
         cmd->constantValues,
         sizeof(cmd->constantValues)
        );

  if (cmd->nmeasurements <= 0 ||
      cmd->nmeasurements > 0xFFFFFF)
  {
    fitting_failed("badMeasurementCount");
    return;
  }

  if (cmd->nparameters <= 0 ||
      cmd->nparameters > ND + NC)
  {
    fitting_failed("badParameterCount");
    return;
  }

  if (!set_measurement_size(cmd->nmeasurements))
  {
    fitting_failed("outOfMemory");
    return;
  }

  nparameters = cmd->nparameters;

  fm = (struct FitterMeasurement*)(cmd + 1);
  memcpy(measurements, fm, sizeof(*fm) * nmeasurements);
  SortMeasurements(0, nmeasurements - 1);
  /* Find the maximum measurement, used for integration... */
  xend = measurements[nmeasurements - 1].bound;
  
  fp = (PRInt32*)(fm + nmeasurements);
  memcpy(param_index, fp, sizeof(PRInt32) * nparameters);

  /* Allocate memory for fitting... */
  if (fitter_e != NULL)
    PR_Free(fitter_e);
  nparamxmeasurements = nparameters * nmeasurements;
  nparam2 = nparameters * nparameters;
  fitter_e = PR_Malloc(
                       (3 * nmeasurements +
                        5 * nparameters +
                        nparam2 +
                        nparamxmeasurements) *
                       sizeof(double)
                      );
  if (fitter_e == NULL)
  {
    fitting_failed("outOfMemory");
    return;
  }
  fitter_hx = fitter_e + nmeasurements;
  fitter_jacTe = fitter_hx + nmeasurements;
  fitter_jac = fitter_jacTe + nparameters;
  fitter_jacTjac = fitter_jac + nparamxmeasurements;
  fitter_Dp = fitter_jacTjac + nparam2;
  fitter_diag_jacTjac = fitter_Dp + nparameters;
  fitter_pDp = fitter_diag_jacTjac + nparameters;
  fitter_wrk = fitter_pDp + nparameters;
  fitter_p = fitter_wrk + nmeasurements;
  fitter_updp = 1;
  fitter_K = imin(10, nparameters);
  fitter_Dp_L2 = DBL_MAX;

  for (i = 0; i < nparameters; i++)
    fitter_p[i] = GET_PARAM(i);

  solution_state = SOLVE_FOR_E;

  EvaluateFunction(fitter_p, fitter_hx);
}

static void
cmd_suspend(void *data, PRUint32 length)
{
  struct FitterReplyHeader irh;
  helper_state = SUSPENDED;
  irh.length = sizeof(struct IntegratorReplyHeader);
  irh.reply = FITTING_HALTED;
  send_packet((char*)&irh, sizeof(irh), PR_FALSE);
}

static void
cmd_stop(void *data, PRUint32 length)
{
  helper_state = DYING;
}

static inline void
integration_done()
{
#ifdef DEBUG_MODE
  printf("integration_done(): %u/%u\n", current_measurement_mark,
         nmeasurements);
#endif
  integrator_call = PR_FALSE;
}

static inline void
fitting_done()
{
  struct FitterReplyHeader *frh;
  PRUint32 l;
  double *params;

  helper_state = SUSPENDED;
  l = sizeof(struct FitterReplyHeader) + sizeof(double) * nparameters;
  frh = PR_Malloc(l);

  frh->length = l;
  frh->reply = FITTING_DONE;

  params = (double*)(frh + 1);
  memcpy(params, fitter_p, sizeof(double) * nparameters);

  if (fitter_e != NULL)
    PR_Free(fitter_e);
  fitter_e = NULL;

  send_packet((char*)frh, l, PR_TRUE);
}

static inline void
add_points(double xval, double *values)
{
  while (
         current_measurement_mark < nmeasurements &&
         measurements[current_measurement_mark].bound <= xval
        )
  {
    double at_x;
    double interp_vals[NS + ND];
    /* There could be several measurements at the same point. To optimise, we
     * should deal with them all in one interpolation...
     */
    at_x = measurements[current_measurement_mark].bound;
    do_interpolation(at_x, interp_vals);
    do
    {
#ifdef DEBUG_MODE
      printf(
             "Measurement %u from %u at %f = %f\n",
             current_measurement_mark,
             measurements[current_measurement_mark].index,
             measurements[current_measurement_mark].bound,
             interp_vals[measurements[current_measurement_mark].index]
            );
#endif
      current_measurements[current_measurement_mark++] =
        interp_vals[measurements[current_measurement_mark].index];
      if (!finite(current_measurements[current_measurement_mark-1]))
        fitter_domain_error = 1;
    }
    while (
           current_measurement_mark < nmeasurements &&
           measurements[current_measurement_mark].bound == at_x
          );
  }
}

inline static void
TRANS_MAT_MAT_MULT(double *a, double *b, PRUint32 n, PRUint32 m, PRUint32 bsize)
{
  PRUint32 i, j, k, jj, kk;
  double sum, *bim, *akm;

  /* compute upper triangular part using blocking */
  for(jj=0; jj<m; jj+=bsize)
  {
    for(i=0; i<m; ++i)
    {
      bim=b+i*m;
      for(j=imax(jj, i); j < imin(jj+bsize, m); ++j)
        bim[j]=0.0; /* b[i*m+j]=0.0; */
    }

    for(kk=0; kk<n; kk+=bsize)
    {
      for(i=0; i<m; ++i)
      {
        bim=b+i*m;
        for(j=imax(jj, i); j < imin(jj+bsize, m); j++)
        {
          sum=0.0;
          for(k=kk; k < imin(kk+bsize, n); k++)
          {
            akm=a+k*m;
            sum+=akm[i]*akm[j]; /* a[k*m+i]*a[k*m+j]; */
          }
          bim[j]+=sum; /* b[i*m+j]+=sum; */
        }
      }
    }
  }

  /* copy upper triangular part to the lower one */
  for(i=0; i<m; ++i)
    for(j=0; j<i; ++j)
      b[i*m+j]=b[j*m+i];
}

/*
 * It would be nice to merge this with do_solve, but for efficiency reasons
 * do_solve assumes the matrix size, so I am using the one from levmar...
 */
/*
 * This function returns the solution of Ax = b
 *
 * The function employs LU decomposition followed by forward/back substitution (see 
 * also the LAPACK-based LU solver above)
 *
 * A is mxm, b is mx1
 *
 * The function returns 0 in case of error,
 * 1 if successfull
 *
 * This function is often called repetitively to solve problems of identical
 * dimensions. To avoid repetitive malloc's and free's, allocated memory is
 * retained between calls and free'd-malloc'ed when not of the appropriate size.
 */
static PRUint32
AX_EQ_B_LU(double *A,
           double *B,
           double *x,
           PRUint32 m)
{
  static char *buf=NULL;
  static PRUint32 buf_sz=0;

  PRInt32 i, j, k;
  PRInt32 *idx, maxi, idx_sz, a_sz, work_sz, tot_sz;
  double *a, *work, max, sum, tmp;

  /* calculate required memory size */
  idx_sz = m;
  a_sz = m*m;
  work_sz = m;
  tot_sz = idx_sz * sizeof(PRInt32) + (a_sz + work_sz) * sizeof(double);

  if (tot_sz > buf_sz)
  {
    /* insufficient memory, allocate a "big" memory chunk at once */
    /* free previously allocated memory */
    if (buf)
      PR_Free(buf);
    buf_sz = tot_sz;
    buf = (char*)PR_Malloc(tot_sz * sizeof(char));
    if (buf == NULL)
    {
      fitting_failed("outOfMemory");
    }
  }

  idx = (PRInt32 *)buf;
  a = (double*)(idx + idx_sz);
  work = a + a_sz;

  /* avoid destroying A, B by copying them to a, x resp. */
  for (i = 0; i < m; i++)
  {
    /* B & 1st row of A */
    a[i] = A[i];
    x[i] = B[i];
  }
  for(  ; i < a_sz; i++)
  {
    /* copy A's remaining rows */
    a[i] = A[i];
  }

  /* compute the LU decomposition of a row permutation of matrix a;
   * the permutation itself is saved in idx[]
   */
  for(i=0; i<m; i++)
  {
    max=0.0;
    for (j=0; j<m; ++j)
      if ((tmp=abs_d(a[i*m+j])) > max)
        max=tmp;
    if(max==0.0)
    {
        return 0;
    }
    work[i]=1.0/max;
  }

  for(j=0; j<m; j++)
  {
    for (i=0; i<j; ++i)
    {
      sum = a[i*m+j];
      for (k=0; k<i; k++)
        sum -= a[i*m+k] * a[k*m+j];
      a[i*m+j] = sum;
    }
    max = 0.0;
    for (i=j; i < m; i++)
    {
      sum = a[i*m+j];
      for (k=0; k<j; ++k)
        sum -= a[i*m+k] * a[k*m+j];
      a[i*m+j]=sum;
      if ((tmp=work[i]*abs_d(sum))>=max)
      {
        max=tmp;
        maxi=i;
      }
    }
    if(j != maxi)
    {
      for(k=0; k<m; k++)
      {
        tmp=a[maxi*m+k];
        a[maxi*m+k]=a[j*m+k];
        a[j*m+k]=tmp;
      }
      work[maxi]=work[j];
    }
    idx[j]=maxi;
    if(a[j*m+j]==0.0)
      a[j*m+j]=EPSILON;
    if(j!=m-1)
    {
      tmp=1.0/(a[j*m+j]);
      for (i=j+1; i<m; i++)
        a[i*m+j]*=tmp;
    }
  }

  /* The decomposition has now replaced a. Solve the linear system using
   * forward and back substitution
   */
  for(i=k=0; i<m; i++)
  {
    j = idx[i];
    sum = x[j];
    x[j] = x[i];
    if (k!=0)
      for (j = k-1; j<i; j++)
        sum -= a[i*m+j] * x[j];
    else if (sum != 0.0)
        k = i+1;
    x[i] = sum;
  }

  for(i = m-1; i >= 0; --i)
  {
    sum = x[i];
    for (j = i+1; j<m; j++)
      sum -= a[i*m+j] * x[j];
    x[i] = sum / a[i*m+i];
  }
  return 1;
}


/*
 * This is the main evaluation routine...
 */
static void
EvaluateFunction(double *params, double *imeasurements)
{
  PRUint32 i;
  current_measurements = imeasurements;
  fitter_domain_error = 0;
  current_measurement_mark = 0;

  /* Set the parameters... */
  for (i = 0; i < nparameters; i++)
    SET_PARAM(i, params[i]);

  memcpy(
         all_yvalues,
         all_y0values,
         sizeof(all_y0values)
        );

  /* Zero out lots of things... */
#define Z(x) memset(x, 0, sizeof(x))
  Z(zz);
  Z(scal);
  Z(ff);
  Z(jac);
  Z(e1);
  Z(e2r);
  Z(e2i);
  Z(cont);
  Z(ip1);
  Z(ip2);
  Z(iphes);
#undef Z
  xpoint = x = start_bound;
  /* Do some initialisations */
  unexp = PR_FALSE;
  unexn = PR_FALSE;
  change = PR_FALSE;
  current_order = MINIMAL_ORDER;
  n_current_order = MAXIMAL_ORDER * ND;
  ikeep = 0;
  ichan = 0;
  theta = 0;
  thetat = 0;
  lrc = n_current_order + ND;
  do_coercv(current_order);
  posneg = ((xend - x) >= 0) ? 1 : -1;
  maximal_step_size = abs_d(xend - x);
  step_size = INITIAL_STEP_SIZE;
  if (step_size > maximal_step_size)
    step_size = maximal_step_size;
  step_size *= posneg;
  step_size_old = step_size;
  reject = PR_FALSE;
  first = PR_TRUE;
  last = PR_FALSE;
  caljac = PR_FALSE;
  skipjac = PR_FALSE;
  calhes = PR_FALSE;
  newton_count = 0;

  /* If they try a very small range, we do it in one step... */
  if ((x + step_size*1.0001 - xend) * posneg > 0.0)
  {
    step_size = xend - x;
    last = PR_TRUE;
  }
  step_size_opt = step_size;
  faccon = 1.0;
  xold = x;
  expo = 1.0 / (current_order + 1.0);
  expmns = (current_order + 1.0)/(2 * current_order);
  quott = abs_tolerance / rel_tolerance;
  rel_tolerance_1 = 0.1 * pow(rel_tolerance, expmns);
  abs_tolerance_1 = rel_tolerance_1 * quott;
  hhfac = step_size;
  nstep = 0;
  naccpt = 0;

  for (i = 0; i < ND; i++)
    scal[i] = abs_tolerance_1 + rel_tolerance_1 * abs_d(all_yvalues[i + NS]);

  /* Add the first point... */
  EvaluateSimple(x, all_yvalues);
  /* Evaluate the differentials into y_0... */
  ComputeDifferentials(x, all_yvalues, y_0);
  
  integrator_call = PR_TRUE;
  do_integrator_cycle();
}

static void
BeginJacIteration()
{
  fitter_d = abs_d((double)(1E-04) * fitter_p[fitter_jac_iteration]);
  fitter_d = max(FITTER_DIFF_DELTA, fitter_d);
#ifdef DEBUG_MODE
  printf("fitter_d = %f\n", fitter_d);
#endif
  fitter_savep = fitter_p[fitter_jac_iteration];
  fitter_p[fitter_jac_iteration] += fitter_d;
  solution_state = COMPUTE_JACOBIAN;
  EvaluateFunction(fitter_p, fitter_wrk);
}

static void
RejectIncrement()
{
  /* if this point is reached, either the linear system could not be solved or
   * the error did not reduce; in any case, the increment must be rejected
   */

  PRUint32 i;
  fitter_mu *= fitter_nu;
  fitter_nu *= 2;

#ifdef DEBUG_MODE
  printf("fitter_mu = %f\n", fitter_mu);
#endif

  for (i = 0; i < nparameters; i++)
    fitter_jacTjac[i * nparameters + i] = fitter_diag_jacTjac[i];

  solution_state = BEGIN_ITERATION;
  return;
}

static void
BeginInnerLoop()
{
  /* Determine increment using adaptive damping...
   * 1) Augment normal equations
   */
  PRUint32 i;
  double tmp;
  for (i=0; i < nparameters; i++)
    fitter_jacTjac[i*nparameters + i] += fitter_mu;

  /* 2) Solve augmented equations... */
  if (AX_EQ_B_LU(fitter_jacTjac, fitter_jacTe, fitter_Dp, nparameters))
  {
    for (i = 0, fitter_Dp_L2 = 0.0; i < nparameters; i++)
    {
#ifdef DEBUG_MODE
      printf("fitter_Dp[%u] = %f\n", i, fitter_Dp[i]);
#endif
      fitter_pDp[i] = fitter_p[i] + (tmp = fitter_Dp[i]);
      fitter_Dp_L2 += tmp * tmp;
    }

    if (fitter_Dp_L2 <= 1E-17 * 1E-17 * (fitter_p_L2 + 1E-17 * 1E-17))
    {
#ifdef DEBUG_MODE
      printf("FITTER_STOP_THRESH_SQ reached(fitter_Dp_L2 = %f).\n",
             fitter_Dp_L2);
#endif
      fitting_done();
      return;
    }

    if (fitter_Dp_L2 >= (fitter_p_L2 + abs_tolerance) / EPSILON2)
    {
      fitting_failed("almostSingular");
      return;
    }
    solution_state = SOLVE_FOR_STEP;
    EvaluateFunction(fitter_pDp, fitter_wrk);
    return;
  }
#ifdef DEBUG_MODE
  printf("System couldn't be solved.");
#endif
  RejectIncrement();
}

static void
BeginIteration()
{
  PRUint32 i, j, l;
  double tmp;
  if ((fitter_updp && fitter_nu > 16) || fitter_updjac == fitter_K)
  {
    fitter_jac_iteration = 0;
    BeginJacIteration();
    return;
  }

  if (fitter_newjac)
  {
    fitter_newjac = 0;
    if (nparamxmeasurements <= BLOCKSZ_SQ)
    {
      /*
       * this is a "small" problem, so we can use the straightforward algorithm
       * rather than worry about cache misses.
       */
      for (i = 0; i < nparameters; i++)
      {
        for (j = i; j < nparameters; j++)
        {
          PRInt32 lm;
          for (l = 0, tmp=0.0; l < nmeasurements; l++)
          {
            lm = l * nparameters;
            tmp += fitter_jac[lm + i] * fitter_jac[lm + j];
          }
          fitter_jacTjac[i * nparameters + j] =
            fitter_jacTjac[j * nparameters + i] = tmp;
        }
        for (l = 0, tmp=0.0; l < nmeasurements; l++)
          tmp += fitter_jac[l * nparameters + i] * fitter_e[l];
        fitter_jacTe[i] = tmp;
      }
    }
    else
    {
      /*
       * this is a "larger" problem, so we can use blocking to avoid cache
       * misses...
       */
      TRANS_MAT_MAT_MULT(fitter_jac, fitter_jacTjac, nmeasurements, nparameters, BLOCKSZ);
      for (i = 0; i < nparameters; i++)
        fitter_jacTe[i] = 0.0;
      for (i = 0; i < nmeasurements; i++)
      {
        double* jacrow = fitter_jac + i * nparameters;
        tmp = fitter_e[i];
        for (l = 0; l < nparameters; l++)
          fitter_jacTe[l] += jacrow[l] * tmp;
      }
    }
    for (i = 0, fitter_p_L2 = fitter_jacTe_inf=0.0;
         i < nparameters;
         i++)
    {
#ifdef DEBUG_MODE
      printf("fitter_jacTe[%u] = %f\n", i, fitter_jacTe[i]);
#endif
      if (fitter_jacTe_inf < (tmp = abs_d(fitter_jacTe[i])))
        fitter_jacTe_inf = tmp;
      /* save diagonal entries so that augmentation can be later canceled */
      fitter_diag_jacTjac[i] = fitter_jacTjac[i * nparameters + i];
      fitter_p_L2 += fitter_p[i] * fitter_p[i];
    }
  }

#ifdef DEBUG_MODE
  printf("Current estimate: ");
  for(i=0; i<nparameters; ++i)
    printf("%.9g ", fitter_p[i]);
  printf("-- errors %.9g %0.9g\n", fitter_jacTe_inf, fitter_p_eL2);
#endif

  /* check for convergence */
  if (fitter_jacTe_inf <= abs_tolerance)
  {
    /* no increment for p in this case */
    fitter_Dp_L2 = 0.0;
#ifdef DEBUG_MODE
    printf("FITTER_STOP_THRESH reached.\n");
#endif
    fitting_done();
    return;
  }

  if (fitter_iteration == 0)
  {
    /* compute initial damping factor */
    for (i = 0, tmp = DBL_MIN; i < nparameters; i++)
      if (fitter_diag_jacTjac[i] > tmp)
        tmp = fitter_diag_jacTjac[i];
#ifdef DEBUG_MODE
    printf("Maximal diagonal = %f\n", tmp);
    printf("Damping factor = %f\n", tmp * FITTER_MU);
#endif
    fitter_mu = FITTER_MU * tmp;
  }

  BeginInnerLoop();
}

static void
SolveForEStep()
{
  double tmp;
  PRUint32 i;

  if (fitter_domain_error)
  {
    fitting_failed("modelInvalidAtStart");
    return;
  }

  for (i = 0, fitter_p_eL2=0.0; i < nmeasurements; i++)
  {
    fitter_e[i] = tmp = measurements[i].value - fitter_hx[i];
#ifdef DEBUG_MODE
    printf("fitter_e[%u] = %f\n", i, tmp);
#endif
    fitter_p_eL2 += tmp * tmp;
  }
  fitter_nu = 20;
  fitter_iteration = 0;

  BeginIteration();
}

static void
ComputeJacobianStep()
{
  PRUint32 i, j;

  if (fitter_domain_error)
  {
    fitting_failed("couldntComputeJac");
    return;
  }


  fitter_p[fitter_jac_iteration] = fitter_savep;
  fitter_d = 1.0 / fitter_d;
  for (i = 0; i < nmeasurements; i++)
    fitter_jac[i * nparameters + fitter_jac_iteration] =
      (fitter_wrk[i] - fitter_hx[i]) * fitter_d;

  fitter_jac_iteration++;
  if (fitter_jac_iteration < nparameters)
  {
    BeginJacIteration();
    return;
  }

#ifdef DEBUG_MODE
  printf("Computed Jacobian...\n");
  for (i = 0; i < nmeasurements; i++)
    for (j = 0; j < nparameters; j++)
      printf("jac[measure=%u,param=%u] = %f\n",
             i, j, fitter_jac[i * nparameters + j]);
#endif

  fitter_nu = 2;
  fitter_updjac = 0;
  fitter_updp = 0;
  fitter_newjac = 1;
  BeginIteration();
}

static void
SolveMainStep()
{
  PRUint32 i, j, l;
  double tmp, dF, dL;

  if (fitter_domain_error)
  {
#ifdef DEBUG_MODE
    printf("fitter_domain_error\n");
#endif
    RejectIncrement();
    return;
  }

  for (i = 0, fitter_pDp_eL2 = 0.0;
       i < nmeasurements;
       i++)
  {
    tmp = measurements[i].value - fitter_wrk[i];
    fitter_pDp_eL2 += tmp * tmp;
  }

  dF = fitter_p_eL2 - fitter_pDp_eL2;
  if (fitter_updp || dF > 0)
  {
    for (i = 0; i < nmeasurements; i++)
    {
      for (l = 0, tmp=0.0; l < nparameters; l++)
        tmp += fitter_jac[i * nmeasurements + l] * fitter_Dp[l];
      tmp = (fitter_wrk[i] - fitter_hx[i] - tmp) / fitter_Dp_L2;
      for (j = 0; j < nparameters; j++)
        fitter_jac[i*nparameters + j] += tmp * fitter_Dp[j];
    }
    fitter_updjac++;
    fitter_newjac = 1;
  }

  for (i = 0, dL=0.0; i < nparameters; i++)
    dL += fitter_Dp[i] * (fitter_mu * fitter_Dp[i] + fitter_jacTe[i]);
  if (dL > 0.0 && dF > 0.0)
  {
    /* reduction in error, increment is accepted */
    dF = 2.0 * dF / dL - 1.0;
    tmp = dF * dF * dF;
    tmp = 1.0 - tmp * tmp * dF;
    fitter_mu = fitter_mu * ( (tmp >= (1/3)) ? tmp : (1.0/3.0));
#ifdef DEBUG_MODE
    printf("After increment, mu=%f\n", fitter_mu);
#endif
    fitter_nu = 2;

    /* update p's estimate */
    for (i=0; i < nparameters; i++)
      fitter_p[i] = fitter_pDp[i];

    /* update e, hx and ||e||_2 */
    for (i=0; i < nmeasurements; i++)
    {
      fitter_e[i] = measurements[i].value - fitter_wrk[i];
      fitter_hx[i] = fitter_wrk[i];
    }

    fitter_p_eL2 = fitter_pDp_eL2;
    fitter_updp = 1;

    fitter_iteration++;
#ifdef FITTER_KMAX
    if (fitter_iteration < FITTER_KMAX)
    {
#endif
      BeginIteration();
      return;
#ifdef FITTER_KMAX
    }
    else
    {
#ifdef DEBUG_MODE
      printf("Iteration limit reached.\n");
#endif
      fitting_done();
      /* fitting_failed("Too many iterations needed."); */
      return;
    }
#endif
  }
#ifdef DEBUG_MODE
  printf("No error reduction: dF=%e dL=%e\n",
         dF, dL);
#endif
  RejectIncrement();
}

static void
do_fitter_cycle()
{
  if (integrator_call)
  {
    do_integrator_cycle();
    if (integrator_call)
      return;
  }
  switch (solution_state)
  {
  case SOLVE_FOR_E:
    SolveForEStep();
    return;
  case COMPUTE_JACOBIAN:
    ComputeJacobianStep();
    return;
  case SOLVE_FOR_STEP:
    SolveMainStep();
    return;
  case BEGIN_ITERATION:
    BeginIteration();
    return;
  }
}
